// file      : xsd/cxx/parser/impl-source.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2007 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx/parser/impl-source.hxx>

#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

namespace CXX
{
  namespace Parser
  {
    namespace
    {
      struct PrintCall: Traversal::Type,

                        Traversal::Fundamental::NameTokens,
                        Traversal::Fundamental::IdRefs,

                        Traversal::Fundamental::QName,

                        Traversal::Fundamental::Base64Binary,
                        Traversal::Fundamental::HexBinary,

                        Traversal::Fundamental::Date,
                        Traversal::Fundamental::DateTime,
                        Traversal::Fundamental::Duration,
                        Traversal::Fundamental::Day,
                        Traversal::Fundamental::Month,
                        Traversal::Fundamental::MonthDay,
                        Traversal::Fundamental::Year,
                        Traversal::Fundamental::YearMonth,
                        Traversal::Fundamental::Time,

                        Context
      {
        PrintCall (Context& c, String const& tag, String const& arg)
            : Context (c), tag_ (tag), arg_ (arg)
        {
        }

        virtual Void
        traverse (SemanticGraph::Type&)
        {
          default_format ();
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::NameTokens& t)
        {
          String type (xs_ns_name () + L"::string_sequence");

          if (default_type (t, type))
            default_format ();
          else
          {
            os << cout_inst << " << " << L << "\"" << tag_ << ": \";"
               << endl
               << "for (" << type << "::const_iterator i (" << arg_ <<
              ".begin ()), e (" << arg_ << ".end ());" << endl
               << "i != e;)"
               << "{"
               << cout_inst << " << *i++;"
               << "if (i != e)" << endl
               << cout_inst << " << " << L << "' ';"
               << "}"
               << cout_inst << " << std::endl;";
          }
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::IdRefs& t)
        {
          String type (xs_ns_name () + L"::string_sequence");

          if (default_type (t, type))
            default_format ();
          else
          {
            os << cout_inst << " << " << L << "\"" << tag_ << ": \";"
               << endl
               << "for (" << type << "::const_iterator i (" << arg_ <<
              ".begin ()), e (" << arg_ << ".end ());" << endl
               << "i != e;)"
               << "{"
               << cout_inst << " << *i++;"
               << "if (i != e)" << endl
               << cout_inst << " << " << L << "' ';"
               << "}"
               << cout_inst << " << std::endl;";
          }
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::QName& t)
        {
          if (default_type (t, xs_ns_name () + L"::qname"))
            default_format ();
          else
            os << cout_inst << " << " << L << "\"" << tag_ << ": \";"
               << endl
               << "if (" << arg_ << ".prefix ().empty ())" << endl
               << cout_inst << " << " << arg_ << ".name ();"
               << "else" << endl
               << cout_inst << " << " << arg_ << ".prefix () << " << L <<
              "':' << " << arg_ << ".name ();"
               << endl
               << cout_inst << " << std::endl;";
        }

        // Binary.
        //
        virtual Void
        traverse (SemanticGraph::Fundamental::Base64Binary& t)
        {
          String type (L"::std::auto_ptr< " + xs_ns_name () + L"::buffer >");

          if (default_type (t, type))
            default_format ();
          else
            os << cout_inst << " << " << L << "\"" << tag_ << ": \" << "
               << arg_ << "->size () << " << L << "\" bytes\" << std::endl;";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::HexBinary& t)
        {
          String type (L"::std::auto_ptr< " + xs_ns_name () + L"::buffer >");

          if (default_type (t, type))
            default_format ();
          else
            os << cout_inst << " << " << L << "\"" << tag_ << ": \" << "
               << arg_ << "->size () << " << L << "\" bytes\" << std::endl;";
        }

        // Date/time.
        //
        virtual Void
        traverse (SemanticGraph::Fundamental::Date& t)
        {
          if (default_type (t, xs_ns_name () + L"::date"))
            default_format ();
          else
            os << cout_inst << " << " << L << "\"" << tag_ << ": \"" << endl
               << " << " << arg_ << ".year () << " << L << "'-'" << endl
               << " << " << arg_ << ".month () << " << L << "'-'" << endl
               << " << " << arg_ << ".day () << "
               << arg_ << ".zone () << std::endl;";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::DateTime& t)
        {
          if (default_type (t, xs_ns_name () + L"::date_time"))
            default_format ();
          else
            os << cout_inst << " << " << L << "\"" << tag_ << ": \"" << endl
               << " << " << arg_ << ".year () << " << L << "'-'" << endl
               << " << " << arg_ << ".month () << " << L << "'-'" << endl
               << " << " << arg_ << ".day () << " << L << "'T'" << endl
               << " << " << arg_ << ".hours () << " << L << "':'" << endl
               << " << " << arg_ << ".minutes () << " << L << "':'" << endl
               << " << " << arg_ << ".seconds () << "
               << arg_ << ".zone () << std::endl;";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::Duration& t)
        {
          if (default_type (t, xs_ns_name () + L"::duration"))
            default_format ();
          else
            os << cout_inst << " << " << L << "\"" << tag_ << ": \";"
               << endl
               << "if (" << arg_ << ".negative ())" << endl
               << cout_inst << " << " << L << "'-';"
               << endl
               << cout_inst << " << " << L << "'P'" << endl
               << " << " << arg_ << ".years () << " << L << "'Y'" << endl
               << " << " << arg_ << ".months () << " << L << "'M'" << endl
               << " << " << arg_ << ".days () << " << L << "\"DT\"" << endl
               << " << " << arg_ << ".hours () << " << L << "'H'" << endl
               << " << " << arg_ << ".minutes () << " << L << "'M'" << endl
               << " << " << arg_ << ".seconds () << " << L << "'S'"
               << " << std::endl;";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::Day& t)
        {
          if (default_type (t, xs_ns_name () + L"::gday"))
            default_format ();
          else
            os << cout_inst << " << " << L << "\"" << tag_ << ": ---\""
               << " << " << arg_ << ".day () << "
               << arg_ << ".zone () << std::endl;";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::Month& t)
        {
          if (default_type (t, xs_ns_name () + L"::gmonth"))
            default_format ();
          else
            os << cout_inst << " << " << L << "\"" << tag_ << ": --\""
               << " << " << arg_ << ".month () << "
               << arg_ << ".zone () << std::endl;";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::MonthDay& t)
        {
          if (default_type (t, xs_ns_name () + L"::gmonth_day"))
            default_format ();
          else
            os << cout_inst << " << " << L << "\"" << tag_ << ": --\"" << endl
               << " << " << arg_ << ".month () << " << L << "'-'" << endl
               << " << " << arg_ << ".day () << "
               << arg_ << ".zone () << std::endl;";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::Year& t)
        {
          if (default_type (t, xs_ns_name () + L"::gyear"))
            default_format ();
          else
            os << cout_inst << " << " << L << "\"" << tag_ << ": \""
               << " << " << arg_ << ".year () << "
               << arg_ << ".zone () << std::endl;";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::YearMonth& t)
        {
          if (default_type (t, xs_ns_name () + L"::gyear_month"))
            default_format ();
          else
            os << cout_inst << " << " << L << "\"" << tag_ << ": \"" << endl
               << " << " << arg_ << ".year () << " << L << "'-'" << endl
               << " << " << arg_ << ".month () << "
               << arg_ << ".zone () << std::endl;";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::Time& t)
        {
          if (default_type (t, xs_ns_name () + L"::time"))
            default_format ();
          else
            os << cout_inst << " << " << L << "\"" << tag_ << ": \"" << endl
               << " << " << arg_ << ".hours () << " << L << "':'" << endl
               << " << " << arg_ << ".minutes () << " << L << "':'" << endl
               << " << " << arg_ << ".seconds () << "
               << arg_ << ".zone () << std::endl;";
        }

      private:
        void
        default_format ()
        {
          os << cout_inst << " << " << L << "\"" << tag_ << ": \" << " <<
            arg_ << " << std::endl;";
        }

        bool
        default_type (SemanticGraph::Type& t, String const& def_type)
        {
          return ret_type (t) != def_type;
        }

      private:
        String tag_;
        String arg_;
      };

      struct Enumeration: Traversal::Enumeration,
                          protected virtual Context
      {
        Enumeration (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          String const& name (eimpl (e));
          String const& ret (ret_type (e));
          SemanticGraph::Type& base (e.inherits ().base ());
          String const& base_ret (ret_type (base));

          os << "// " << name << endl
             << "//" << endl
             << endl;

          // pre
          //
          os << "void " << name << "::" << endl
             << "pre ()"
             << "{"
             << "}";

          // post
          //
          os << ret << " " << name << "::" << endl
             << post_name (e) << " ()"
             << "{";

          if (ret == base_ret)
          {
            os << (ret != L"void" ? "return " : "") <<
              post_name (base) << " ();";
          }
          else if (ret == L"void")
          {
            os << arg_type (base) << " v (" << post_name (base) << " ());"
               << endl;

            if (options.value<CLI::generate_print_impl> ())
            {
              PrintCall t (*this, e.name (), "v");
              t.dispatch (base);
            }
            else
              os << "// TODO" << endl
                 << "//" << endl;
          }
          else
          {
            if (base_ret == L"void")
              os << post_name (base) << " ();";
            else
              os << arg_type (base) << " v (" << post_name (base) << " ());"
                 << endl
                 << "// TODO" << endl
                 << "//" << endl
                 << "// return ... ;" << endl;
          }

          os << "}";
        }
      };

      //
      //
      struct List: Traversal::List, protected virtual Context
      {
        List (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& l)
        {
          String const& name (eimpl (l));
          SemanticGraph::Type& type (l.argumented ().type ());

          String item (unclash (ename (l), "item"));

          os << "// " << name << endl
             << "//" << endl
             << endl;

          // pre
          //
          os << "void " << name << "::" << endl
             << "pre ()"
             << "{"
             << "}";

          // item
          //
          String const& arg (arg_type (type));

          os << "void " << name << "::" << endl
             << item;

          if (arg == L"void")
            os << " ();";
          else
            os << " (" << arg << " " << item << ")";

          os << "{";

          if (arg != L"void")
          {
            if (options.value<CLI::generate_print_impl> ())
            {
              PrintCall t (*this, type.name (), item);
              t.dispatch (type);
            }
            else
              os << "// TODO" << endl
                 << "//" << endl;
          }

          os << "}";

          // post
          //
          String const& ret (ret_type (l));

          os << ret << " " << name << "::" << endl
             << post_name (l) << " ()"
             << "{";

          if (ret != L"void")
            os << "// TODO" << endl
               << "//" << endl
               << "// return ... ;" << endl;

          os << "}";
        }
      };

      //
      //
      struct Union: Traversal::Union, protected virtual Context
      {
        Union (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& u)
        {
          String const& name (eimpl (u));

          os << "// " << name << endl
             << "//" << endl
             << endl;

          // pre
          //
          os << "void " << name << "::" << endl
             << "pre ()"
             << "{"
             << "}";

          // _characters
          //
          os << "void " << name << "::" << endl
             << "_characters (const " << string_type << "& s)"
             << "{";

          if (options.value<CLI::generate_print_impl> ())
            os << cout_inst << " << " << L << "\"" << u.name () << ": \"" <<
              " << s << std::endl;";
          else
            os << "// TODO" << endl
               << "//" << endl;

          os << "}";

          // post
          //
          String const& ret (ret_type (u));

          os << ret << " " << name << "::" << endl
             << post_name (u) << " ()"
             << "{";

          if (ret != L"void")
            os << "// TODO" << endl
               << "//" << endl
               << "// return ... ;" << endl;

          os << "}";
        }
      };

      //
      //
      struct ParserHook: Traversal::Member,
                         protected virtual Context
      {
        ParserHook (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
	  if (skip (m))
            return;

          String const& name (ename (m));
          String const& arg (arg_type (m.type ()));

          os << "void " <<
            eimpl (dynamic_cast<SemanticGraph::Complex&> (m.scope ())) <<
            "::" << endl
             << name;

          if (arg == L"void")
            os << " ()";
          else
            os << " (" << arg << " " << name << ")";

          os << "{";

          if (arg != L"void")
          {
            if (options.value<CLI::generate_print_impl> ())
            {
              PrintCall t (*this, m.name (), name);
              t.dispatch (m.type ());
            }
            else
              os << "// TODO" << endl
                 << "//" << endl;
          }

          os << "}";
        }
      };

      //
      //
      struct Complex : Traversal::Complex,
                       protected virtual Context
      {
        Complex (Context& c)
            : Context (c), parser_hook_ (c)
        {
          names_parser_hook_ >> parser_hook_;
        }

        virtual Void
        traverse (Type& c)
        {
          String const& name (eimpl (c));

          Boolean restriction (restriction_p (c));

          os << "// " << name << endl
             << "//" << endl
             << endl;

          // pre
          //
          os << "void " << name << "::" << endl
             << "pre ()"
             << "{"
             << "}";

          // Parser hooks.
          //
          if (!restriction_p (c))
            names (c, names_parser_hook_);

          // post
          //
          String const& ret (ret_type (c));

          os << ret << " " << name << "::" << endl
             << post_name (c) << " ()"
             << "{";

          if (c.inherits_p ())
          {
            SemanticGraph::Type& base (c.inherits ().base ());
            String const& base_ret (ret_type (base));

            if (ret == base_ret)
            {
              os << (ret != L"void" ? "return " : "") <<
                post_name (base) << " ();";
            }
            else if (ret == L"void")
            {
              os << arg_type (base) << " v (" << post_name (base) << " ());"
                 << endl;

              if (options.value<CLI::generate_print_impl> ())
              {
                PrintCall t (*this, c.name (), "v");
                t.dispatch (base);
              }
              else
                os << "// TODO" << endl
                   << "//" << endl;
            }
            else
            {
              if (base_ret == L"void")
                os << post_name (base) << " ();";
              else
                os << arg_type (base) << " v (" << post_name (base) << " ());"
                   << endl
                   << "// TODO" << endl
                   << "//" << endl
                   << "// return ... ;" << endl;
            }
          }
          else
          {
            if (ret != L"void")
              os << "// TODO" << endl
                 << "//" << endl
                 << "// return ... ;" << endl;
          }

          os << "}";
        }

      private:
        //
        //
        ParserHook parser_hook_;
        Traversal::Names names_parser_hook_;
      };
    }

    Void
    generate_impl_source (Context& ctx)
    {
      if (ctx.options.value<CLI::generate_print_impl> ())
        ctx.os << "#include <iostream>" << endl
               << endl;

      Traversal::Schema schema;

      Traversal::Sources sources;
      Traversal::Names schema_names;

      Namespace ns (ctx);
      Traversal::Names names;

      schema >> sources >> schema;
      schema >> schema_names >> ns >> names;

      List list (ctx);
      Union union_ (ctx);
      Complex complex (ctx);
      Enumeration enumeration (ctx);

      names >> list;
      names >> union_;
      names >> complex;
      names >> enumeration;

      schema.dispatch (ctx.schema_root);
    }
  }
}

